// Copyright 2009 Intel Corporation
// SPDX-License-Identifier: Apache-2.0

#include <cstdlib>
#include <memory>
#include "../common.h"
#include "../utility/AbstractArray.h"
#include "Fabric.h"

namespace rkcommon {
  namespace networking {
    Fabric::Fabric() {}
  }
}

